#include "pow.h"

extern XColor	 lut_colorcell_defs[256];
extern PictColorTable *PowColorTable;

int PowSetupColormap(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[] ) {
  /* This routine creates a toplevel window with a colormap for that
     window which has the specified number of cells free in addition to those
     needed by POW.  If force_cmap is non-zero:

     0 - Default behavior.  Choose the "best" pseudocolor map with enough
         colorcells available.

     1 - Force a new private pseudocolor colormap.

     2 - Force truecolor mode.  Note, the *images* are truecolor, the visual
         associated with the POW colormap may be anything,
	  "toplevel -visual best" is used to choose the visual.
   
     3 - Force screen default colormap (WARNING: this may crash application
         if the default colormap is pseudocolor, or not if you're lucky, or 
         VISU may barf at image startup, you getting the idea that I don't
         recommend this option; if it's truecolor, this should work as well 
         as option 2 does. )

     Returns toplevel path name.
     All of an application's windows should then use the colormap of this
     toplevel. */
  Tk_Window tkwin,dotwin;
  Screen *screen;
  Display *disp;
  Colormap cmap;
  int i;
  XColor *colors;
  int colormap_size;
  int screenIndex;
  unsigned long *pixels;
  unsigned long *plane_masks;
  Window dotwinID;
  Visual *visual;
  XVisualInfo *visual_info;
  char *toplevel;
  char *options;
  int free_cells;
  int ncolors;
  Status status;
  Bool tfGotColors;        /* True if we got colors from X */
  int force_cmap; 
  int powCells;

  if (argc == 2 && (!strcmp(argv[1],"none") || !strcmp(argv[1],"NULL"))) {
    /* do nothing */
    return TCL_OK;
  }



  if (argc > 5 || argc < 3 ) {
    interp->result =
      "usage: powSetupColormap toplevel_name free_cells ?force_cmap? ?options_list?";
    return TCL_ERROR;
  }


  toplevel = ckalloc(strlen(argv[1])+5);
  strcpy(toplevel,argv[1]);

  

  Tcl_GetInt(interp,argv[2],&free_cells);

  if (argc >= 4) {
    Tcl_GetInt(interp,argv[3],&force_cmap);
  } else {
    force_cmap = 0;
  }



  if (argc == 5) {
    options = ckalloc(strlen(argv[4])+1);
    strcpy(options,argv[4]);
  } else {
    options = ckalloc(1);
    *options = '\0';
  }
#if defined(__WIN32__) || defined(macintosh)
  /*In WIN32/mac, just create the toplevel. Pseudocolor mode isn't supported
    in WIN32/mac because the necessary Xlib calls are missing.*/
  Tcl_SetVar(interp,"powPseudoImages","0",TCL_GLOBAL_ONLY); 
  return Tcl_VarEval(interp, "toplevel ",toplevel,
		       " -visual best ",options,(char *)NULL);
#else

  if (force_cmap == 2) {
    Tcl_SetVar(interp,"powPseudoImages","0",TCL_GLOBAL_ONLY); 
    return Tcl_VarEval(interp, "toplevel ",toplevel,
		     " -visual best ",options,(char *)NULL);    
  }

  /* find out some things about the screen*/

  dotwin = Tk_NameToWindow(interp,".",Tk_MainWindow(interp));
  screen=Tk_Screen(dotwin);
  disp = Tk_Display(dotwin);
  screenIndex=DefaultScreen(disp);

  if (force_cmap == 3) {
    /*Force default visual. */

    return Tcl_VarEval(interp, "toplevel ",toplevel," -visual default ", 
		       options,(char *)NULL);

    /* We want to know if the default visual is pseudocolor. 
       Finding this out in Xlib is stupidly complicated, so we'll
       let Tk do it for us */
    Tcl_Eval(interp,"string match \"pseudocolor\" [winfo screenvisual .]");
    if (interp->result != "1") {
      Tcl_SetVar(interp,"powPseudoImages","0",TCL_GLOBAL_ONLY); 
    }
    return TCL_OK;
  }
  
  visual_info = get_visual(disp);  

  if (visual_info == NULL) {
    (void)fprintf (stderr,"powSetupColormap: toplevel created with non-pseudocolor visual. \n");
    Tcl_SetVar(interp,"powPseudoImages","0",TCL_GLOBAL_ONLY); 
    return Tcl_VarEval(interp, "toplevel ",toplevel," -visual best ", 
		       options,(char *)NULL);
  }


  plane_masks = (unsigned long *) ckalloc(MAXPLANES*sizeof(unsigned long));
  pixels = (unsigned long *) ckalloc(MAX_COLORS*sizeof(unsigned long));
  if (plane_masks == NULL || pixels == NULL) {
    (void)fprintf(stderr,
		  "\n Unable to allocate storage for PowSetupColormap\n");
    return TCL_ERROR;
  }


  tfGotColors = False;           /* Will be set True when we succeed */

  if (force_cmap != 1) {
  
    /* Try default Colormap first */
    cmap = DefaultColormap(disp, screenIndex); 
    
    ncolors = 212; /* why not 256, I don't know, but since we have to
		      match whatever VISU does and I don't want to
		      mess with that.... */
    
    while (ncolors > 10) {
      status = XAllocColorCells(disp,cmap,True,plane_masks,0,pixels,ncolors);
      if (status != 0) {
	tfGotColors = True;   /* Success.  Break out of the while loop */
	break;
	
      }
      ncolors -= 10;      /* Failure.  Decrement request and try again */
    } /* End of while loop */
  }
    
 Tcl_GetInt(interp,Tcl_GetVar(interp,"powMinColorcells",TCL_GLOBAL_ONLY),
	    &powCells);

  if(force_cmap != 1  && ncolors >= powCells + free_cells) {  /* enough colors , use default map */
    /* First free the colors we got */      
    XFreeColors (disp, cmap, pixels, ncolors, 0);
    ckfree((void*)plane_masks);  plane_masks = NULL;
    ckfree((void*)pixels);       pixels = NULL;
    return Tcl_VarEval(interp, "toplevel ",toplevel,
		       options,(char *)NULL);
  } else {         /* not enough colors */
    
    /* Free the colors we did get */
    if(tfGotColors == True) {
      XFreeColors (disp, cmap, pixels, ncolors, 0);
    }
    
    
    tkwin = Tk_CreateWindowFromPath(interp,dotwin,".powCmap",NULL);
    

    if (tkwin == NULL) {
      (void)fprintf(stderr,
		    "\n Couldn't create dummy window for PowSetupColormap\n");
      return TCL_ERROR;
    }

    /*  This prevents a seg-fault from occuring at a 'winfo class' command  */
    Tk_SetClass( tkwin, "PowCmapDmy" );
    
    
    colormap_size = DisplayCells(disp,screenIndex);
    colors = (XColor*)ckalloc(colormap_size*sizeof(XColor));
    
    for(i=0;i<colormap_size;i++) {
      pixels[i] = colors[i].pixel = i;
      colors[i].flags = DoRed | DoGreen | DoBlue;
    }
    
    XQueryColors(disp,DefaultColormap(disp, screenIndex),
		 colors,colormap_size);
    
    
    cmap = XCreateColormap (disp, RootWindow(disp,screenIndex),
			    visual_info->visual , AllocNone); 
    
    if (!cmap) {
      printf ("ERROR in PowSetupColormap: XCreateColormap returned %x\n", 
	      (unsigned int)cmap);
      return TCL_ERROR;
    }
    
    
    colormap_size = 212 - powCells - free_cells;
    XAllocColorCells(disp,cmap,True,plane_masks,0,pixels,colormap_size); 
    XStoreColors(disp,cmap,colors,colormap_size);
    

    ckfree((void*)plane_masks);  plane_masks = NULL;
    ckfree((void*)pixels);       pixels = NULL;
    ckfree((void*)colors);

    Tk_SetWindowColormap(tkwin,cmap);
  
    return Tcl_VarEval(interp, "toplevel ",toplevel,
		       " -colormap .powCmap ",options,(char *)NULL);
  }
#endif /* __WIN32__ || macintosh */
}





int PowTestColormap(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[] ) {
  /* This routine returns the number of free colors in the colormap of
     the specified window */
#if defined(__WIN32__) || defined(macintosh)
  /* This routine does nothing in WIN32 or MacOS */
  return TCL_OK;
#else
  Tk_Window tkwin;
  Display *disp;
  Colormap cmap;
  unsigned long *pixels;
  unsigned long *plane_masks;
  char *window;
  int ncolors;
  Status status;
  Bool tfGotColors;        /* True if we got colors from X */

  if (argc != 2 ) {
    interp->result =
      "usage: powTestColormap window";
    return TCL_ERROR;
  }


  window = ckalloc(strlen(argv[1])+1);
  strcpy(window,argv[1]);


  plane_masks = (unsigned long *) ckalloc(MAXPLANES*sizeof(unsigned long));
  pixels = (unsigned long *) ckalloc(MAX_COLORS*sizeof(unsigned long));
  if (plane_masks == NULL || pixels == NULL) {
    (void)fprintf(stderr,
		  "\n Unable to allocate storage for PowTestColormap\n");
    return TCL_ERROR;
  }


  /* find out some things about the screen*/

  tkwin = Tk_NameToWindow(interp,window,Tk_MainWindow(interp));
  disp = Tk_Display(tkwin);

  tfGotColors = False;
  
  cmap = Tk_Colormap(tkwin);
  
  ncolors = 256;


  while (ncolors > 0) {
    status = XAllocColorCells(disp,cmap,True,plane_masks,0,pixels,ncolors);
    if (status != 0) {
      tfGotColors = True;   /* Success.  Break out of the while loop */
      break;

    }
    ncolors -= 1;      /* Failure.  Decrement request and try again */
  } /* End of while loop */
  
  
    
  /* First free the colors we got */      
  if (tfGotColors == True) {
    XFreeColors (disp, cmap, pixels, ncolors, 0);
  }
  ckfree((void*)plane_masks);  plane_masks = NULL;
  ckfree((void*)pixels);       pixels = NULL;
  sprintf(interp->result,"%i",ncolors);
  return TCL_OK;
#endif /* __WIN32__ || macintosh */
}


int PowSetupPhotoImages(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[] ) {
  int ncolors=256;
  int lut_start=0;
  int color_def;

  /*We're going to use the Colortable routines in VISU to setup our colortables
    and fake out the X stuff */

  for (color_def=lut_start; color_def<lut_start+ncolors;
       color_def++) {
    lut_colorcell_defs[color_def].pixel = color_def;
    lut_colorcell_defs[color_def].flags = (DoRed | DoGreen | DoBlue);
  }


  AllocateColorTable((PictColorTable **)&PowColorTable,(Display*)NULL,
		     (Colormap)0L,POW_COLORMAP,
		     ncolors,lut_start,0);
  return TCL_OK;
}
  

  


void PowDitherToPhoto(PowImage *powImage, Tk_PhotoImageBlock *photoBlock,
		      double disp_min, double disp_max) 
{
  /* Dithers the powImage data to a Tk_PhotoBlock.  Calls 
     convert_block_to_byte and uses lut_colorcell_defs global. */
  unsigned char *dithered_data_array;
  unsigned char *photo_image_array;
  unsigned char *inptr;
  unsigned char *outptr;
  void *in_data;
  int width,height;
  int data_type;
  int *red,*green,*blue;
  int *intensity_lut,*red_lut,*green_lut,*blue_lut;
  int col,i,j;
  char *elem;

  width = powImage->width;
  height = powImage->height;

  in_data = (powImage->dataptr)->data_array;

  data_type = (powImage->dataptr)->data_type;

  dithered_data_array = (unsigned char *)ckalloc(sizeof(char)*width*height);

  convert_block_to_byte(in_data,dithered_data_array,width*height,data_type,
			&disp_min, &disp_max);

  photo_image_array = (unsigned char *)ckalloc(3*sizeof(char)*width*height);

  inptr = dithered_data_array;
  outptr = photo_image_array;

  for (i=0; i<height; i++) {
    inptr = dithered_data_array + i*width;
    outptr = photo_image_array + (height - i - 1) * width * 3;
    for ( j = 0; j < width; j++) {
      col = *inptr++;
      *outptr++ = lut_colorcell_defs[col].red   >> 8;
      *outptr++ = lut_colorcell_defs[col].green >> 8;
      *outptr++ = lut_colorcell_defs[col].blue  >> 8;
    }
  }


  ckfree(dithered_data_array);
  
  photoBlock->pixelPtr = photo_image_array;

  return;
    
}

int PowPhotoColorTable (ClientData clientData, Tcl_Interp *interp,
			int argc, char *argv[])
{
  void (*f)(Display *display,Colormap cmap,
	    int ncolors,int lut_start,char overlay,
	    int *red,int *green,int *blue,
	    int *intensity_lut,int *red_lut,int *green_lut,
	    int *blue_lut);
  int length;

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"",
		     argv[0], " cmap\"", (char *) NULL);
    return TCL_ERROR;
  }

  if(strcmp(argv[1], "gray") == 0)
    f = gray;
  else if(strcmp(argv[1], "blkbdy") == 0)
    f = blkbdy;
  else if(strcmp(argv[1], "hot") == 0)
    f = hot;
  else if(strcmp(argv[1], "cold") == 0)
    f= cold;
  else if(strcmp(argv[1], "hls") == 0)
    f = hls;
  else if(strcmp(argv[1], "rgb") == 0)
    f = rgb;
  else if(strcmp(argv[1], "invert") == 0)
    f = invert_cmap;
  else if(strcmp(argv[1], "random") == 0)
    f = randwalk_spectrum;
  else if(strcmp(argv[1], "bowlerhat") == 0)
    f = bowlerhat;
  else if(strcmp(argv[1], "tophat") == 0)
    f = tophat;
  else if(strcmp(argv[1], "hatgray") == 0)
    f = hatgray;
  else if(strcmp(argv[1], "hatct") == 0)
    f = hatct;
  else if(strcmp(argv[1], "gray-ramp2") == 0)
    f = gray_ramp2;
  else if(strcmp(argv[1], "gray-ramp4") == 0)
    f = gray_ramp4;
  else if(strcmp(argv[1], "gray-step4") == 0)
    f = gray_step4;
  else if(strcmp(argv[1], "gray-step8") == 0)
    f = gray_step8;
  else if(strcmp(argv[1], "bgr-step") == 0)
    f = bgr_step;
  else if(strcmp(argv[1], "bgr-ramp") == 0)
    f = bgr_ramp;
  else if(strcmp(argv[1], "bgr-step2") == 0)
    f = bgr_step2;
  else if(strcmp(argv[1], "bgr-ramp2") == 0)
    f = bgr_ramp2;
  else if(strcmp(argv[1], "rygcbm-ramp") == 0)
    f = rygcbm_ramp;
  else if(strcmp(argv[1], "rygcbm-step") == 0)
    f = rygcbm_step;
  else if(strcmp(argv[1], "spectrum") == 0)
    f = spectrum2;
  else if(strcmp(argv[1], "inv_spec") == 0)
    f = inv_spec;
  else if(strcmp(argv[1], "color1") == 0)
    f = color1_lut;
  else if(strcmp(argv[1], "color2") == 0)
    f = color2_lut;
  else if(strcmp(argv[1], "color3") == 0)
    f = color3_lut;

  else {
     /*  Try to find a user-specified cmap...

	 ... well, not yet.

     char scrtch[200];
     Tcl_Obj *lut;

     sprintf(scrtch,"cmapLUT_%s,powDef",argv[1]);
     lut = Tcl_ObjGetVar2(interp,
			  Tcl_NewStringObj("powImageParam",-1),
			  Tcl_NewStringObj(scrtch,-1),
			  TCL_GLOBAL_ONLY);
     if( lut==NULL ) {
	sprintf(scrtch,"Unable to locate LUT for %s\n",argv[1]);
	Tcl_SetResult(interp,scrtch,TCL_VOLATILE);
	return TCL_ERROR;
     }
     */

     return TCL_ERROR;
  }
  
  (*f)((Display*)NULL,
       (Colormap)0L,
       PowColorTable->ncolors,
       PowColorTable->lut_start,
       False,
       PowColorTable->red,PowColorTable->green,PowColorTable->blue,
       PowColorTable->intensity_lut,
       PowColorTable->red_lut,PowColorTable->green_lut,
       PowColorTable->blue_lut);
  return TCL_OK;
}




int PowPhotoCmapStretch (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[] ) {
  int cwid,clen;
  int x_lut[50];
  int y_lut[50];
  int i,j;
  int lut_size;
  int *p_lut;
  
  if (argc < 5) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
			     " cwid clen x1 y1 x2 y2 ... ]\"", 
			     (char *) NULL);
    return TCL_ERROR;
  } else {
    p_lut = PowColorTable->intensity_lut;
    lut_size = PowColorTable->ncolors;
    if((Tcl_GetInt(interp, argv[1], &cwid) != TCL_OK) ||
       (Tcl_GetInt(interp, argv[2], &clen) != TCL_OK) ) {
      Tcl_AppendResult(interp, "bad lookup table : should be \"", argv[0],
		       "  cwid clen x1 y1 x2 y2 ... \"", 
		       (char *) NULL);
      return TCL_ERROR;
    }
    i = 3;
    j=0;
    while(i<argc-1) {
      if((Tcl_GetInt(interp, argv[i++], &x_lut[j]) != TCL_OK) ||
	 (Tcl_GetInt(interp, argv[i++], &y_lut[j]) != TCL_OK) ) {
	Tcl_AppendResult(interp, "bad lookup table : should be \"", argv[0],
			 " cwid clen x1 y1 x2 y2 ... \"", 
			 (char *) NULL);
	return TCL_ERROR;
      }
      j++;
    }
    
    for(i=0;i<j;i++) {
      x_lut[i] = floor((double)x_lut[i]/(double)cwid*lut_size); 
      y_lut[i] = floor((double)(clen-y_lut[i])/
		       (double)clen*lut_size);
    }
    
    non_linear_lut(p_lut,lut_size,
		   x_lut,y_lut,j,
		   (Display*)NULL,
		   (Colormap)0L,
		   PowColorTable->ncolors,
		   PowColorTable->lut_start,
		   False,
		   PowColorTable->red,PowColorTable->green,PowColorTable->blue,
		   PowColorTable->intensity_lut,
		   PowColorTable->red_lut,PowColorTable->green_lut,
		   PowColorTable->blue_lut); 
    
    return TCL_OK;
  }
}


int PowReditherPhotoBlock (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[] ) {
  PowImage *image_instance;
  Tk_PhotoHandle photo_handle;
  Tk_PhotoImageBlock photo_block;
  double min, max;

  if (argc != 4) {
    interp->result = "usage: powReditherPhotoBlock imageName min max";
    return TCL_ERROR;
  }

#if TK_MAJOR_VERSION < 8
  if ((photo_handle = Tk_FindPhoto(argv[1])) == NULL) {
#else
  if ((photo_handle = Tk_FindPhoto(interp,argv[1])) == NULL) {
#endif    
    Tcl_AppendResult(interp, "image \"", argv[1], "\" doesn't exist",
		     (char *) NULL);
    return TCL_ERROR;
  }

  Tcl_GetDouble(interp, argv[2], &min);
  Tcl_GetDouble(interp, argv[3], &max);

  /*  Tk_PhotoGetImage(photo_handle, &photo_block); */

  image_instance = PowFindImage(argv[1]);

  PowDitherToPhoto(image_instance, &photo_block, min, max);

  photo_block.pixelSize = 3;
  photo_block.width = image_instance->width;
  photo_block.height = image_instance->height;
  photo_block.pitch = image_instance->width * 3;
  photo_block.offset[0] = 0;
  photo_block.offset[1] = 1;
  photo_block.offset[2] = 2;
  
  
  Tk_PhotoPutBlock(photo_handle,&photo_block,0,0,image_instance->width,image_instance->height);
  
  ckfree(photo_block.pixelPtr);
  return TCL_OK;
}

