#include "pow.h"
#include <limits.h>

/* on some system , e.g. linux, SUNs DBL_MAX is in float.h */
#ifndef DBL_MAX
#include <float.h>
#endif

#ifndef DBL_MIN
#include <float.h>
#endif


int PowFindCurvesMinMax(char *curves, char *axis, double *min, double *max,
			int filter)
  /* Sets min and max to the minimum and maximum values found in the 
     "axis" vector of each member of the list of curves.  Initial values 
     for min and max are whatever they are when the routine is called.  
     (i.e. if *min == 0  upon entry and there are no negative values in the 
     curves, *min will == 0 upon return). */
{
  PowCurve *current_curve;
  PowVector *current_vector;
  PowData *current_data;
  int curve_index,curveArgc;
  char **curveArgv;
  double datum;
  int i;

  if(curves == NULL || strstr(curves,"NULL") != NULL ) return TCL_OK;
  if(Tcl_SplitList(interp,curves,&curveArgc,&curveArgv) != TCL_OK) {
    fprintf(stderr, "%s\n", interp->result);
    return TCL_ERROR;
  }
  for (curve_index = 0; curve_index < curveArgc;curve_index++) {
    current_curve = PowFindCurve(curveArgv[curve_index]);
    switch (*axis) {
    case 'X': current_vector = current_curve->x_vector;
      break;
    case 'Y': current_vector = current_curve->y_vector;
      break;
    case 'Z': current_vector = current_curve->z_vector;
      break;
    default: sprintf(interp->result, "Axis must be X, Y, or Z.");
      return TCL_ERROR;
    }
    if(current_vector != NULL) {
      current_data = current_vector->dataptr;
      for (i = current_vector->offset ;
	   i < current_vector->offset + current_curve->length ;
	   i++) {
	datum = PowExtractDatum(current_data,i);
	if( filter && datum <= 0.0 ) /* Positive data only (for logs) */
	   datum = DBL_MAX;
	if (datum != DBL_MAX) {
	  *min = (datum < *min) ? datum : *min;
	  *max = (datum > *max) ? datum : *max;
	}
      }
    } else {
       *min = (         1.0          < *min) ?          1.0          : *min;
       *max = (current_curve->length > *max) ? current_curve->length : *max;
    }
  }
  ckfree((char *) curveArgv);
  return TCL_OK;
}

int PowFindCurvesBBox(char *graph,
		      char *curves, double *xleft, double *xright, 
		      double *ybot, double *ytop, WCSdata *BaseWCS)
{
   PowCurve *current_curve;
   PowVector *Xvec, *Yvec;
   PowData *current_data;
   int curve_index,curveArgc;
   char **curveArgv;
   double xdatum, ydatum, xmin, xmax, ymin, ymax, tmp;
   double lxmin, lxmax, lymin, lymax;
   int i,j, logX, logY;
   char *optVal;

   if(curves == NULL || strstr(curves,"NULL") != NULL ) return TCL_OK;

   if(Tcl_SplitList(interp,curves,&curveArgc,&curveArgv) != TCL_OK) {
      fprintf(stderr, "%s\n", interp->result);
      return TCL_ERROR;
   }

   for( curve_index = 0; curve_index < curveArgc; curve_index++ ) {

      current_curve = PowFindCurve(curveArgv[curve_index]);

      Xvec = current_curve->x_vector;
      Yvec = current_curve->y_vector;

      optVal = PowGetObjectOption( graph, curveArgv[curve_index],
				   "logX", "curve" );
      if( !optVal || Tcl_GetBoolean( interp, optVal, &logX )==TCL_ERROR ) {
	 logX = 0;
      }
      optVal = PowGetObjectOption( graph, curveArgv[curve_index],
				   "logY", "curve" );
      if( !optVal || Tcl_GetBoolean( interp, optVal, &logY )==TCL_ERROR ) {
	 logY = 0;
      }

      xmin = ymin = DBL_MAX;  xmax = ymax = - DBL_MAX;
      if( current_curve->WCS.WCSexists==0 || Xvec==NULL || Yvec==NULL ) {

	 lxmin = lymin = DBL_MAX;  lxmax = lymax = - DBL_MAX;
	 PowFindCurvesMinMax( curveArgv[curve_index], "X", &lxmin, &lxmax,
			      logX );
	 PowFindCurvesMinMax( curveArgv[curve_index], "Y", &lymin, &lymax,
			      logY );

	 if( logX ) {
	    if( lxmin<=0.0 || lxmax<=0.0 ) {
	       return TCL_ERROR;
	    } else {
	       lxmin = log10(lxmin);
	       lxmax = log10(lxmax);
	    }
	 }
	 if( logY ) {
	    if( lymin<=0.0 || lymax<=0.0 ) {
	       return TCL_ERROR;
	    } else {
	       lymin = log10(lymin);
	       lymax = log10(lymax);
	    }
	 }
	 xmin = (lxmin < xmin) ? lxmin : xmin;
	 xmax = (lxmax > xmax) ? lxmax : xmax;
	 ymin = (lymin < ymin) ? lymin : ymin;
	 ymax = (lymax > ymax) ? lymax : ymax;

	 if( PowPixToPos( xmin, ymin, &current_curve->WCS, &xmin, &ymin ) )
	    return TCL_ERROR;
	 if( PowPixToPos( xmax, ymax, &current_curve->WCS, &xmax, &ymax ) )
	    return TCL_ERROR;

	 if( PowPosToPix( xmin, ymin, BaseWCS, &xmin, &ymin ) )
	    return TCL_ERROR;
	 if( PowPosToPix( xmax, ymax, BaseWCS, &xmax, &ymax ) )
	    return TCL_ERROR;

	 if( xmin>xmax ) { tmp=xmax; xmax=xmin; xmin=tmp; }
	 if( ymin>ymax ) { tmp=ymax; ymax=ymin; ymin=tmp; }

      } else {

	 for (i = Xvec->offset, j = Yvec->offset ;
	      i < Xvec->offset + current_curve->length ;
	      i++, j++) {
	    xdatum = PowExtractDatum(Xvec->dataptr,i);
	    ydatum = PowExtractDatum(Yvec->dataptr,j);
	    if( xdatum != DBL_MAX && ydatum != DBL_MAX ) {

	       if( PowPixToPos( xdatum-1, ydatum-1, &current_curve->WCS,
				&xdatum, &ydatum ) )
		  return TCL_ERROR;
	       if( PowPosToPix( xdatum, ydatum, BaseWCS, &xdatum, &ydatum ) )
		  return TCL_ERROR;
	       xmin = (xdatum < xmin) ? xdatum : xmin;
	       xmax = (xdatum > xmax) ? xdatum : xmax;
	       ymin = (ydatum < ymin) ? ydatum : ymin;
	       ymax = (ydatum > ymax) ? ydatum : ymax;
	    }
	 }
      }

      if( xmin < *xleft  ) *xleft  = xmin;
      if( ymin < *ybot   ) *ybot   = ymin;
      if( xmax > *xright ) *xright = xmax;
      if( ymax > *ytop   ) *ytop   = ymax;

   }

   ckfree((char *) curveArgv);
   return TCL_OK;
}

int PowFindImagesBBox(char *images, double *xleft, double *xright, 
		      double *ybot, double *ytop, WCSdata *BaseWCS)
{
  PowImage *current_image;
  int image_index,imageArgc, maxDim;
  char **imageArgv;
  double xorigin,xotherend,yorigin,yotherend,tmp;

  if(images == NULL || strstr(images,"NULL") != NULL) return TCL_OK;


  if(Tcl_SplitList(interp,images,&imageArgc,&imageArgv) != TCL_OK) {
    fprintf(stderr, "%s\n", interp->result);
    return TCL_ERROR;
  }

  for (image_index = 0; image_index < imageArgc; image_index++) {

     current_image = PowFindImage(imageArgv[image_index]);
    
     /*  Convert origin and otherend info into pixel coordinates  */

     if ( PowPosToPix( current_image->xorigin, current_image->yorigin,
		       BaseWCS, &xorigin, &yorigin ) )
	return TCL_ERROR;
     if ( PowPosToPix( current_image->xotherend, current_image->yotherend,
		       BaseWCS, &xotherend, &yotherend ) )
	return TCL_ERROR;

     /**************************************/
     /*  Test the images for consistency:  */
     /**************************************/

     if ( BaseWCS->WCSexists && !current_image->WCS.WCSexists ) {
	Tcl_SetResult(interp,"All images do not contain WCS information",
		      TCL_VOLATILE);
	return TCL_ERROR;
     }

     /*  We are in pixel coordinates, so they should  */
     /*  ALWAYS go from left->right                   */

     if ( xorigin > xotherend || yorigin > yotherend ) {
	Tcl_SetResult(interp,"All images do not point in same direction",
		      TCL_VOLATILE);
	return TCL_ERROR;
     }

     /*  Images must have "nearly" identical rotations  */

     tmp = ( current_image->width > current_image->height ?
	     current_image->width : current_image->height ) + 1;
     tmp = 57.3 * atan( 1.0 / tmp );
     if ( BaseWCS->WCSexists &&
	  fabs( BaseWCS->rot - current_image->WCS.rot ) > tmp ) {
	sprintf(interp->result, "Rotation angle of images don't match.");
	return TCL_ERROR;
     }

     /**************************************/
     /*     End of consistency tests       */
     /**************************************/

     if ( xorigin   < *xleft  ) *xleft   = xorigin;
     if ( yorigin   < *ybot   ) *ybot    = yorigin;
     if ( xotherend > *xright ) *xright  = xotherend;
     if ( yotherend > *ytop   ) *ytop    = yotherend;

  }

  ckfree(imageArgv);
  return TCL_OK;
}

int PowSortGraphMinMax( PowGraph *graph, double *xleft, double *xright,
			double *ybot, double *ytop, double *xdim, double *ydim)
{
  double tmp;

  if ( PowPosToPix(*xleft,  *ybot, &graph->WCS, xleft,  ybot) )
     return TCL_ERROR;
  if ( PowPosToPix(*xright, *ytop, &graph->WCS, xright, ytop) )
     return TCL_ERROR;

  if ( *xleft>*xright ) { tmp=*xleft; *xleft=*xright; *xright=tmp; }
  if ( *ybot >*ytop )   { tmp=*ybot;  *ybot =*ytop;   *ytop  =tmp; }

  *xdim = *xright - *xleft;
  *ydim = *ytop   - *ybot;

/*  Convert bounds back into graph coordinates and return  */

  if ( PowPixToPos(*xleft,  *ybot, &graph->WCS, xleft,  ybot) )
     return TCL_ERROR;
  if ( PowPixToPos(*xright, *ytop, &graph->WCS, xright, ytop) )
     return TCL_ERROR;

  return TCL_OK;
}

PowCurve *
PowFindCurve(char *curve_name) {
  Tcl_HashEntry *entry_ptr;
  PowCurve *curve_ptr;

  if(curve_name == NULL || strstr(curve_name,"NULL") != NULL) {
    return (PowCurve *) NULL;
  }
  
  entry_ptr = Tcl_FindHashEntry(&PowCurveTable,curve_name);
  if (entry_ptr == NULL) {
    return (PowCurve *) NULL;
  }
  curve_ptr = (PowCurve *) Tcl_GetHashValue(entry_ptr);
  return curve_ptr;
}

PowImage *
PowFindImage(char *image_name) {
  Tcl_HashEntry *entry_ptr;
  PowImage *image_ptr;

  if(image_name == NULL || strstr(image_name,"NULL") != NULL) {
    return (PowImage *) NULL;
  }
  
  entry_ptr = Tcl_FindHashEntry(&PowImageTable,image_name);
  if (entry_ptr == NULL) {
    return (PowImage *) NULL;
  }
  image_ptr = (PowImage *) Tcl_GetHashValue(entry_ptr);
  return image_ptr;
}


double
PowExtractDatum(PowData *data, int element) {
  double datum;
  switch (data->data_type) {
  case BYTE_DATA : datum = (double) *((unsigned char *) data->data_array + element);
    if (datum == UCHAR_MAX) { datum = DBL_MAX;}
    break;
  case INT_DATA : datum = (double) *((int *) data->data_array + element);
    if (datum == INT_MAX) {datum = DBL_MAX;}
    break;
  case SHORTINT_DATA : datum = (double) *((short int *) data->data_array + element);
    if (datum == SHRT_MAX) {datum = DBL_MAX;}
    break;
  case REAL_DATA : datum = (double) *((float *) data->data_array + element);
    if (datum == FLT_MAX) {datum = DBL_MAX;}
    break;	
  case DOUBLE_DATA : datum =  *((double *) data->data_array + element);
    break;	
  case STRING_DATA : /*don't use PowExtractDatum on string data */
    datum = DBL_MAX;    
    break;
  }
  return datum;
}


int 
PowPutDatum(PowData *data, double datum, int element) {
  switch (data->data_type) {
  case BYTE_DATA : *((unsigned char *) data->data_array + element) = 
		     (unsigned char) datum;
    break;
  case INT_DATA : *((int *) data->data_array + element) =
		    (int) datum;
    break;
  case SHORTINT_DATA : *((short int *) data->data_array + element) =
			 (int) datum;
    break;
  case REAL_DATA : *((float *) data->data_array + element) =
		    (float) datum;
    break;	
  case DOUBLE_DATA :  *((double *) data->data_array + element) =
		    (double) datum;
    break;	
  }
  return TCL_OK;
}



PowVector *
PowFindVector(char *vector_name) {
  Tcl_HashEntry *entry_ptr;
  PowVector *vector_ptr;

  if(vector_name == NULL || strstr(vector_name,"NULL") != NULL) {
    return (PowVector *) NULL;
  }
  
  entry_ptr = Tcl_FindHashEntry(&PowVectorTable,vector_name);
  if (entry_ptr == NULL) {
    return (PowVector *) NULL;
  }
  vector_ptr = (PowVector *) Tcl_GetHashValue(entry_ptr);
  return vector_ptr;
}

PowData *
PowFindData(char *data_name) {
  Tcl_HashEntry *entry_ptr;
  PowData *data_ptr;

  if(data_name == NULL || strstr(data_name,"NULL") != NULL) {
    return (PowData *) NULL;
  }
  
  entry_ptr = Tcl_FindHashEntry(&PowDataTable,data_name);
  if (entry_ptr == NULL) {
    return (PowData *) NULL;
  }
  data_ptr = (PowData *) Tcl_GetHashValue(entry_ptr);
  return data_ptr;
}

PowGraph *
PowFindGraph(char *graph_name) {
  Tcl_HashEntry *entry_ptr;
  PowGraph *graph_ptr;

  if(graph_name == NULL || strstr(graph_name,"NULL") != NULL) {
    return (PowGraph *) NULL;
  }
  
  entry_ptr = Tcl_FindHashEntry(&PowGraphTable,graph_name);
  if (entry_ptr == NULL) {
    return (PowGraph *) NULL;
  }
  graph_ptr = (PowGraph *) Tcl_GetHashValue(entry_ptr);
  return graph_ptr;
}


char *PowGetObjectOption(char *graph, char *obj, char *option, char *objType)
{
   char *idxStr, gn[255];
   char *res;
   int len;

   len = strlen(graph);

   if( len>5 && !strcmp(graph+len-5,"scope") ) {
      strncpy(gn,graph,len-5);
      gn[len-5]='\0';
   } else {
      strcpy(gn,graph);
   }

   len    = strlen(gn)+strlen(obj)+strlen(option)+10;
   idxStr = (char *) ckalloc( len*sizeof(char) );

   if( !strcmp(objType,"curve") ) {

      sprintf(idxStr,"%s%s,%s",option,obj,gn);
      res = Tcl_GetVar2(interp,"powCurveParam",idxStr,TCL_GLOBAL_ONLY);
      if( res==NULL ) {
	 sprintf(idxStr,"%s,powDef",option);
	 res = Tcl_GetVar2(interp,"powCurveParam",idxStr,TCL_GLOBAL_ONLY);
      }

   } else if( !strcmp(objType,"image") ) {

      sprintf(idxStr,"%s%s,%s",option,obj,gn);
      res = Tcl_GetVar2(interp,"powImageParam",idxStr,TCL_GLOBAL_ONLY);
      if( res==NULL ) {
	 sprintf(idxStr,"%s,powDef",option);
	 res = Tcl_GetVar2(interp,"powImageParam",idxStr,TCL_GLOBAL_ONLY);
      }

   } else if( !strcmp(objType,"graph") ) {

      sprintf(idxStr,"%s%s,%s",option,obj,gn);
      res = Tcl_GetVar2(interp,"powPlotParam",idxStr,TCL_GLOBAL_ONLY);
      if( res==NULL ) {
	 sprintf(idxStr,"%s,powDef",option);
	 res = Tcl_GetVar2(interp,"powPlotParam",idxStr,TCL_GLOBAL_ONLY);
      }

   }
   ckfree(idxStr);
   return res;
}

int PowPosToPix( double  xpos, double  ypos, WCSdata *WCS,
		 double *xpix, double *ypix )
{
   if ( WCS->WCSexists ) {
      if ( pow_xypx( xpos, ypos, WCS->xref, WCS->yref, WCS->xrefpix,
		     WCS->yrefpix, WCS->xinc, WCS->yinc, WCS->rot,
		     WCS->type, xpix, ypix) != 0 ) {
	 sprintf(interp->result,"%s","Couldn't translate WCS coords to pixels");
	 return TCL_ERROR;
      }
   } else {
      *xpix = (xpos - WCS->xref) / WCS->xinc + WCS->xrefpix;
      *ypix = (ypos - WCS->yref) / WCS->yinc + WCS->yrefpix;
   }
   return TCL_OK;
}

int PowPixToPos( double  xpix, double  ypix, WCSdata *WCS,
		 double *xpos, double *ypos )
{
   if ( WCS->WCSexists ) {
      if ( pow_worldpos( xpix, ypix, WCS->xref, WCS->yref, WCS->xrefpix,
			 WCS->yrefpix, WCS->xinc, WCS->yinc, WCS->rot,
			 WCS->type, xpos, ypos) != 0 ) {
	 sprintf(interp->result,"%s","Couldn't translate pixels to WCS coords");
	 return TCL_ERROR;
      }
   } else {
      *xpos = (xpix - WCS->xrefpix) * WCS->xinc + WCS->xref;
      *ypos = (ypix - WCS->yrefpix) * WCS->yinc + WCS->yref;
   }
   return TCL_OK;
}

